/*  This file is part of the simple Vector Field Plotter (vfplot) project

    Written by Nicolas Brodu <nicolas.brodu@numerimoire.net>

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

#ifndef FAST_ATOF_NEXT_TOKEN
#define FAST_ATOF_NEXT_TOKEN

// Usage: for (char* x = line; *x!=0;) {value = fast_atof_next_token(x); ... }
// only classic notation supported, no fancy hex or the like that atof can handle
inline double fast_atof_next_token(char* &str) {
    double value = 0;
    double neg = 1;
    if (*str==0) return 0;
    while ((*str==' ')||(*str=='\t')||(*str=='\n')||(*str=='\r')) {
        ++str; if (*str==0) return 0;
    }
    for (;;++str) {
        switch(*str) {
            default: ++str;// break on invalid characters
            case 0: return value*neg; // end of string
            case '-': neg = -1; continue;
            case '+': continue;
            case '0': value *= 10; continue;
            case '1': value = value * 10 + 1; continue;
            case '2': value = value * 10 + 2; continue;
            case '3': value = value * 10 + 3; continue;
            case '4': value = value * 10 + 4; continue;
            case '5': value = value * 10 + 5; continue;
            case '6': value = value * 10 + 6; continue;
            case '7': value = value * 10 + 7; continue;
            case '8': value = value * 10 + 8; continue;
            case '9': value = value * 10 + 9; continue;
            case '.': {
                double tenpow = 10;
                ++str; if (*str==0) return value*neg; // useless terminal .
                while ((*str>='0')&&(*str<='9')) {
                    value += (*str - '0') / tenpow;
                    tenpow *= 10;
                    ++str; if (*str==0) return value*neg;
                } 
                // ignore unknown characters other than e or E
                if (*str!='e'&&*str!='E') {
                    while ((*str==' ')||(*str=='\t')||(*str=='\n')||(*str=='\r')) ++str;
                    return value*neg;
                }
            }
            case 'e':
            case 'E': {
                int exponum = 0;
                bool div = false;
                for (++str;;++str) {
                    switch(*str) {
                        default: ++str;
                        case 0: {
                            // non-recursive fast-exponentiation
                            // TODO: IEEE754 tricks... dependent of double
                            double expoval = 1;
                            double tenpow = 10;
                            while (exponum!=0) {
                                if ((exponum&1)!=0) expoval *= tenpow;
                                tenpow *= tenpow;
                                exponum /= 2;
                            }
                            if (div) return value*neg/expoval; return value*neg*expoval;
                        }
                        case '+': div = false; continue;
                        case '-': div = true; continue;
                        case '0': exponum *= 10; continue;
                        case '1': exponum = exponum * 10 + 1; continue;
                        case '2': exponum = exponum * 10 + 2; continue;
                        case '3': exponum = exponum * 10 + 3; continue;
                        case '4': exponum = exponum * 10 + 4; continue;
                        case '5': exponum = exponum * 10 + 5; continue;
                        case '6': exponum = exponum * 10 + 6; continue;
                        case '7': exponum = exponum * 10 + 7; continue;
                        case '8': exponum = exponum * 10 + 8; continue;
                        case '9': exponum = exponum * 10 + 9; continue;
                    }
                }
                // shall never reach this point
                if (div) return value*neg;
                return value*neg;
            }
        }
    }
    // shall never reach this point
    return value*neg;
}

#endif
